# cells-rtcproto

A demonstration program for the `rtcproto` (run-to-completion protothreads) library.

The program aims to visually and temporally demonstrate the point of scheduling protothreads cooperatively based on ending timestamps issued by the scheduler. In the background, a predator-prey cellular automata is running at a few steps per second. If you feed a new line on standard input by pressing `Enter`, a protothread will be queued that will take many thousands of steps, taking the simulation to a very distant point.

However, this program is also meant to demonstrate how RTC scheduled protothreads can be combined with double buffering. The cellular automata you are seeing is always the one in the active/front buffer, and the protothread that takes many steps always operates on the passive/back buffer. That's why once we complete the main computation (taking many steps in the cellular automata), we also have to queue an operation to swap the front and back buffers, and to copy the results from the new front buffer to the new back buffer so that the buffers remain synchronized.

The point is that while the thousands of steps are calculated, you can see the program firmly stay in sync with it's main loop, always returning to the beginning of the loop with very little delay. The demo program does this in a crude way, but this return can be made very accurate by combining two things:

- requesting the scheduler to exit a tiny bit before the actual point in time you wish to regain execution,
- synchronizing with or waiting until the expected timepoint with a combination of nanosleep and a very short busy wait period.

## Statistics

The statistics printed below the cellular automata are the following:

- the number of times the scheduler scheduled (entered) the `trophs_bigstep` protothread in total
- the length of time spent executing the protothread in total
- the line number of the local continuation that caused the longest overrun
- the length of the maximal overrun time (beyond the timepoint requested by the scheduler)
- size/capacity/dynamic-target-capacity of: the priority queue of waiting calls, the data structures associated with protothreads, and the local state storage for the three protothreads (bigstep, swap, copy)

## Dynamic capacity

The size/capacity/dynamic-target-capacity statistics are meant to show how the scheduler can let go of allocated but underused capacity slowly over time. If you have very frequent, regular bursts of scheduled work in your system, this helps avoid wasting too much performance on allocating/releasing memory (if you manage to tick the scheduler slowly enough).

## Maximal overrun

The maximal overrun of each protothread is depedent on how eagerly (often) it cooperates with the scheduler, checking whether it's allocated time is up. The `trophs_bigstep` protothread only checks this every couple iterations, you can try reducing this number, and you should see the maximal overrun decrease.
